# startup.py
# ----------
# This file is loaded on startup of FoxDot. You can edit it to include any code
# you want to run at the beginning of a session. To use an alternate startup
# file use the --startup flag followed by a valid path. You can also run FoxDot
# with a --no-startup flag to not load this file.